import 'dart:async';

import 'package:postgres/postgres.dart';
import 'package:uuid/uuid.dart';

class PoolOverloadedException extends Error {}

class LockTimeoutException extends Error {}

abstract class QueryAble {
  Future<int> execute(String sql,
      {Map<String, dynamic>? substitutionValues, int? timeoutInSeconds});
  Future<PostgreSQLResult> query(String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds});
  Future<List<Map<String, Map<String, dynamic>>>> mappedResultsQuery(
      String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds});
  void cancelTransaction({String? reason});
}

class TransactionQeury implements QueryAble {
  PostgreSQLExecutionContext ctx;
  TransactionQeury(this.ctx);

  @override
  Future<int> execute(String sql,
      {Map<String, dynamic>? substitutionValues, int? timeoutInSeconds}) {
    return ctx.execute(sql,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  Future<List<Map<String, Map<String, dynamic>>>> mappedResultsQuery(
      String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds}) {
    return ctx.mappedResultsQuery(fmtString,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  Future<PostgreSQLResult> query(String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds}) {
    // TODO: implement query
    return ctx.query(fmtString,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  void cancelTransaction({String? reason}) {
    ctx.cancelTransaction(reason: reason);
  }
}

typedef TransactionBlock = Future Function(QueryAble ctx);

typedef RemoveConnection = void Function(String cid);

class ConnectionWrapper implements QueryAble {
  bool idle = false;
  int idleTs = 0;
  bool handle = false;
  String cid;
  ConnectionSetting settings;
  RemoveConnection removeMe;
  late PostgreSQLConnection connection;
  late Timer timeoutTimer;
  ConnectionWrapper(this.cid, this.settings, this.removeMe) {
    connection = PostgreSQLConnection(
        settings.host, settings.port, settings.dbName,
        username: settings.userName, password: settings.password);
    timeoutTimer = Timer.periodic(Duration(seconds: 60), (timer) {
      final ts = DateTime.now().millisecondsSinceEpoch;
      if (ts - idleTs > 3600000) {
        timer.cancel();
        connection.close().then((value) {
          removeMe(cid);
        });
      }
    });
  }

  Future open() async {
    await connection.open();
  }

  void lock() {
    handle = true;
  }

  void release() {
    handle = false;
  }

  Future transaction(TransactionBlock block) async {
    await connection.transaction((ctx) async {
      final queryAble = TransactionQeury(ctx);
      await block(queryAble);
    });
  }

  @override
  Future<int> execute(String sql,
      {Map<String, dynamic>? substitutionValues, int? timeoutInSeconds}) {
    idleTs = DateTime.now().millisecondsSinceEpoch;
    return connection.execute(sql,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  Future<List<Map<String, Map<String, dynamic>>>> mappedResultsQuery(
      String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds}) {
    idleTs = DateTime.now().millisecondsSinceEpoch;
    return connection.mappedResultsQuery(fmtString,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  Future<PostgreSQLResult> query(String fmtString,
      {Map<String, dynamic>? substitutionValues,
      bool? allowReuse,
      int? timeoutInSeconds}) {
    idleTs = DateTime.now().millisecondsSinceEpoch;
    return connection.query(fmtString,
        substitutionValues: substitutionValues,
        timeoutInSeconds: timeoutInSeconds);
  }

  @override
  void cancelTransaction({String? reason}) {
    throw UnimplementedError();
  }
}

class ConnectionSetting {
  String host;
  int port;
  String dbName;
  String userName;
  String password;
  ConnectionSetting(
      {required this.host,
      required this.port,
      required this.dbName,
      required this.userName,
      required this.password});
}

class Database {
  Map<String, ConnectionWrapper> pools;
  late ConnectionSetting settings;
  int _poolSize = 0;
  bool lock = false;

  Database._privateConstructor() : pools = <String, ConnectionWrapper>{};

  static final Database _instance = Database._privateConstructor();
  factory Database() {
    return _instance;
  }

  void init(ConnectionSetting config, {int poolSize = 4}) {
    settings = config;
    _poolSize = poolSize;
  }

  Future requireLock() async {
    for (var i = 0; i < 100; i++) {
      if (lock) {
        await Future.delayed(Duration(milliseconds: 5));
      } else {
        return;
      }
    }
    throw LockTimeoutException();
  }

  void releaseLock() {
    lock = false;
  }

  void removeConnection(String connId) {
    pools.remove(connId);
  }

  Future<ConnectionWrapper> getConnection() async {
    await requireLock();

    final idleConnections =
        pools.values.where((conn) => !conn.handle).toList(growable: false);
    if (idleConnections.isNotEmpty) {
      idleConnections.first.lock();
      releaseLock();
      return idleConnections.first;
    } else {
      if (pools.length < _poolSize) {
        final conn = ConnectionWrapper(
            Uuid().v4().toString(), settings, removeConnection);
        conn.lock();
        await conn.open();
        pools.addEntries([MapEntry(conn.cid, conn)]);
        releaseLock();
        return conn;
      } else {
        // TODO: 连接池不为空，又没有空闲的连接。就等待 x ms 后再尝试
        for (var i = 0; i < 100; i++) {
          await Future.delayed(Duration(milliseconds: 10));
          final idleConns = pools.values
              .where((conn) => !conn.handle)
              .toList(growable: false);
          if (idleConns.isNotEmpty) {
            idleConns.first.lock();
            releaseLock();
            return idleConns.first;
          }
        }
        releaseLock();
        throw PoolOverloadedException();
      }
    }
  }
}
